// (C) Copyright 2009 Donald A. Graft, All Rights Reserved

#define NVGLOBAL

#define _WIN32_WINDOWS 0x0500
#include <windows.h>
#include "resource.h"
#include "stdio.h"
#include "CUVIDServer.h"

#define AVC 0
#define VC1 1
#define MP1 2
#define MP2 3
int decode_mode;

LRESULT CALLBACK SelectProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

HANDLE hNALUPumpThread = NULL;
HANDLE hDisplayThread = NULL;
HANDLE hStopNALUPump;
HANDLE hStopDisplay;
HANDLE hStep;
HDC hDC;
HWND hwnd;

char FileName[2048];

extern FILE *bits;
extern int IsFirstByteStreamNALU;
extern void InitializeBuffer(void);
extern int GetAnnexbNALU (unsigned int *nalu_len, unsigned char *buffer);

DWORD WINAPI DisplayThread(LPVOID lpParam) 
{
	HANDLE hArray[2] = {hStopDisplay, hFrameAvailable};
	HANDLE hArray2[2] = {hStopDisplay, hStep};
	extern void ShowFrame(int convert);
	DWORD val;

	for (;;)
	{
		// Ask the server to decode the next frame and wait for it.
		// Allow for cancelling the wait when the user hits the "Reset"
		// button.
		::SetEvent(hFrameRequested);
		if ((val=::WaitForMultipleObjects(2, hArray, FALSE, INFINITE)) == WAIT_OBJECT_0)
			break;
		// Display the frame.
		ShowFrame(1);
		// Wait for the user to hit the "Step" button again.
		// Allow for cancelling the wait when the user hits the "Reset Decoder"
		// button.
		if ((val=::WaitForMultipleObjects(2, hArray2, FALSE, INFINITE)) == WAIT_OBJECT_0)
			break;
	}
	ExitThread(0);
}

DWORD WINAPI NALUPumpThread(LPVOID lpParam) 
{
	int count;
	unsigned int *m = (unsigned int *)(&pMsg->rpc_message[0]);
	HANDLE hArray[2] = {hStopNALUPump, hDone};

	// Open the input file.
	fopen_s(&bits, FileName, "rb");
	if (bits == NULL)
	{
		MessageBox(hwnd, "Failed to open input file.", "Client", MB_OK);
		ExitThread(0);
	}

	m[0] = OP_SET_MODE;
	m[1] = 0;
	m[2] = 1;
	m[3] = 0;
	m[4] = decode_mode;
	m[5] = 0;
	m[6] = 0;
	m[7] = 0;
	m[8] = 0;
	m[9] = 0;
	m[10] = 0;

	::ResetEvent(hDone);
	::SetEvent(hExec);
	// Signal the server to reset.
	::SetEvent(hStopDecoding);
	count = 0;
	while (::WaitForSingleObject(hStopDecoding, 0) == WAIT_OBJECT_0)
	{
		Sleep(100);
		if (++count > 30)
		{
			MessageBox(hwnd, "Server did not respond to our reset request!", "Client", MB_OK);
			break;
		}
	}

	// Initialize NALU feeding.
	IsFirstByteStreamNALU = 1;
	_fseeki64(bits, 0, SEEK_SET);
	InitializeBuffer();
	pMsg->client_process_id = ::GetCurrentProcessId();
	for (;;)
	{
		// Give the server a NALU to decode.
		m[0] = OP_DECODE_NALU;
		if (GetAnnexbNALU(&m[1], (unsigned char *) &m[2]) < 0)
		{
			// Reached end of file.
			// Tell the server to flush the last remaining frames in
			// its DPB.
			m[0] = OP_FLUSH;
			::ResetEvent(hDone);
			::SetEvent(hExec);
			// Wait for the server to finish with it.
			// Allow for cancelling the wait on a reset.
			WaitForMultipleObjects(2, hArray, FALSE, INFINITE);
			// No more data, so stop the NALU pump.
			break;
		}
		::ResetEvent(hDone);
		::SetEvent(hExec);
		// Wait for the server to finish with it.
		// Allow for cancelling the wait on a reset.
		if (WaitForMultipleObjects(2, hArray, FALSE, INFINITE) == WAIT_OBJECT_0)
			break;
	}
	// Close the input file and exit.
	fclose(bits);
	bits = NULL;
    ExitThread(0); 
} 

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG  msg;    
	WNDCLASS wc = {0};
	HICON hIcon;
	HANDLE hMap;
	RECT wrect, crect;
	int Edge_Width, Edge_Height;
	extern void InitYUV2RGB(void);

	InitYUV2RGB();

	wc.lpszClassName = "Application";
	wc.hInstance     = hInstance;
	wc.hbrBackground = GetSysColorBrush(COLOR_3DFACE);
	wc.lpfnWndProc   = WndProc;
	RegisterClass(&wc);

	hwnd = CreateWindow(wc.lpszClassName,"CUVID Client 2.0.0 beta 1",
						WS_OVERLAPPEDWINDOW | WS_VISIBLE | WS_CLIPCHILDREN | WS_MAXIMIZEBOX,
						100, 100, 640, 480, 0, 0, hInstance, 0);  
	GetWindowRect(hwnd, &wrect);
	GetClientRect(hwnd, &crect);
	Edge_Width = wrect.right - wrect.left - crect.right + crect.left;
	Edge_Height = wrect.bottom - wrect.top - crect.bottom + crect.top;
	MoveWindow(hwnd, wrect.left, wrect.top, 640+Edge_Width, 480+Edge_Height, true);

	hIcon = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ICON)); 
	SendMessage(hwnd, (UINT) WM_SETICON, (WPARAM) ICON_SMALL, (LPARAM) hIcon); 
	SendMessage(hwnd, (UINT) WM_SETICON, (WPARAM) ICON_BIG, (LPARAM) hIcon);
	hDC = GetDC(hwnd);

	// Open the shared memory used for communicating with the server.
	hMap = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0,
							   sizeof(TSharedMemory),
							   "NVSharedMem");
	if (hMap == NULL)
	{
		MessageBox(hwnd, "Could not open shared memory!", "Client", MB_OK);
		PostQuitMessage(0);
		return 0;
	}
	if (GetLastError() != ERROR_ALREADY_EXISTS)
	{
		MessageBox(hwnd, "Server not running!", "Client", MB_OK);
		CloseHandle(hMap);
		PostQuitMessage(0);
		return 0;
	}
	pMsg = (TSharedMemory*) ::MapViewOfFile(hMap, FILE_MAP_WRITE, 0, 0, sizeof(TSharedMemory));
	if (pMsg == NULL)
	{
		CloseHandle(hMap);
		PostQuitMessage(0);
		return 0;
	}
	// Open events used for communicating with the server,
	hExec = ::CreateEvent(NULL, FALSE, FALSE, "NVExec");
	hDone = ::CreateEvent(NULL, FALSE, FALSE, "NVDone");
	hStopDecoding = ::CreateEvent(NULL, TRUE, FALSE, "NVStopDecoding");
	hFrameRequested = ::CreateEvent(NULL, FALSE, FALSE, "NVRequested");
	hFrameAvailable = ::CreateEvent(NULL, FALSE, FALSE, "NVAvailable");
	// Open events for local control. They are unnamed to avoid
	// collisions with other clients.
	hStopNALUPump = ::CreateEvent(NULL, FALSE, FALSE, NULL);
	hStopDisplay = ::CreateEvent(NULL, FALSE, FALSE, NULL);
	hStep = ::CreateEvent(NULL, FALSE, FALSE, NULL);

	while (GetMessage(&msg, NULL, 0, 0))
		DispatchMessage(&msg);

	UnmapViewOfFile(pMsg);
	CloseHandle(hMap);
	CloseHandle(hExec);
	CloseHandle(hDone);
	CloseHandle(hStopDecoding);
	CloseHandle(hStopNALUPump);
	CloseHandle(hStopDisplay);
	CloseHandle(hStep);
	CloseHandle(hFrameRequested);
	CloseHandle(hFrameAvailable);

	return (int) msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static HWND hwndReset, hwndPlay, hwndAVCFile, hwndVC1File, hwndMP1File, hwndMP2File;
	PAINTSTRUCT ps;
	unsigned int *m = (unsigned int *)(&pMsg->rpc_message[0]);
	HANDLE aThread[] = {hNALUPumpThread, hDisplayThread};
	DWORD count;
	extern void ShowFrame(int convert);
	int PopFileDlg(PTSTR pstrFileName, HWND hOwner);

	switch(msg)  
	{
	case WM_CREATE:
		hwndAVCFile = CreateWindow("button", "AVC File",
			WS_CHILD | WS_VISIBLE,
			8, 8, 64, 24, hwnd, (HMENU) ID_SELECT_AVC, NULL, NULL);
		hwndVC1File = CreateWindow("button", "VC1 File",
			WS_CHILD | WS_VISIBLE,
			78, 8, 64, 24, hwnd, (HMENU) ID_SELECT_VC1, NULL, NULL);
		hwndMP1File = CreateWindow("button", "MP1 File",
			WS_CHILD | WS_VISIBLE,
			148, 8, 64, 24, hwnd, (HMENU) ID_SELECT_MP1, NULL, NULL);
		hwndMP2File = CreateWindow("button", "MP2 File",
			WS_CHILD | WS_VISIBLE,
			218, 8, 64, 24, hwnd, (HMENU) ID_SELECT_MP2, NULL, NULL);
		hwndReset = CreateWindow("button", "Reset",
			WS_CHILD | WS_VISIBLE,
			288, 8, 64, 24, hwnd, (HMENU) ID_RESET, NULL, NULL);
		hwndPlay = CreateWindow("button", "Step",
			WS_CHILD | WS_VISIBLE,
			358, 8, 64, 24, hwnd, (HMENU) ID_PLAY, NULL, NULL);
		break;

	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
			case ID_SELECT_AVC:
				decode_mode = AVC;
				PopFileDlg(FileName, hwnd);
				break;
			case ID_SELECT_VC1:
				decode_mode = VC1;
				PopFileDlg(FileName, hwnd);
				break;
			case ID_SELECT_MP1:
				decode_mode = MP1;
				PopFileDlg(FileName, hwnd);
				break;
			case ID_SELECT_MP2:
				decode_mode = MP2;
				PopFileDlg(FileName, hwnd);
				break;
			case ID_RESET:
				// Stop local threads.
				::SetEvent(hStopNALUPump);
				::SetEvent(hStopDisplay);
				::SetEvent(hStep);
				::WaitForMultipleObjects(2, aThread, TRUE, 3000);
				::ResetEvent(hStopNALUPump);
				::ResetEvent(hStopDisplay);
				::ResetEvent(hStep);
				// Signal the server to reset.
				::SetEvent(hStopDecoding);
				count = 0;
				while (::WaitForSingleObject(hStopDecoding, 0) == WAIT_OBJECT_0)
				{
					Sleep(100);
					if (++count > 30)
					{
						MessageBox(hwnd, "Server did not respond to our reset request!", "Client", MB_OK);
						break;
					}
				}
				break;
			case ID_PLAY:
				// If decoding is not already running start it by creating the
				// NALU pump and display threads, otherwise set the step event to
				// display the next frame.
step:
				if (hNALUPumpThread == NULL || ::WaitForSingleObject(hNALUPumpThread, NULL) == WAIT_OBJECT_0)
				{
					hNALUPumpThread = CreateThread(NULL, 0, NALUPumpThread, NULL, 0, NULL);
					Sleep(300);
					if (hDisplayThread == NULL || ::WaitForSingleObject(hDisplayThread, NULL) == WAIT_OBJECT_0)
					{
						hDisplayThread = CreateThread(NULL, 0, DisplayThread, NULL, 0, NULL);
					}
				}
				else
				::SetEvent(hStep);
				break;
		}
		break;

	case WM_MOUSEWHEEL:
		if ((short) HIWORD(wParam) < 0)
			goto step;
		break;

	case WM_PAINT:
		{
			HDC hdc;

			if (hNALUPumpThread)
			{
				hdc = BeginPaint(hwnd, &ps);
				ShowFrame(0);
				EndPaint(hwnd, &ps);
			}
			break;
		}

	case WM_DESTROY:
		// Stop local threads.
		::SetEvent(hStopNALUPump);
		::SetEvent(hStopDisplay);
		::SetEvent(hStep);
		::WaitForMultipleObjects(2, aThread, TRUE, 3000);
		::ResetEvent(hStopNALUPump);
		::ResetEvent(hStopDisplay);
		::ResetEvent(hStep);
		// Signal the server to reset.
		::SetEvent(hStopDecoding);
		count = 0;
		while (::WaitForSingleObject(hStopDecoding, 0) == WAIT_OBJECT_0)
		{
			Sleep(100);
			if (++count > 30)
			{
				MessageBox(hwnd, "Server did not respond to our reset request!", "Client", MB_OK);
				break;
			}
		}
		PostQuitMessage(0);
		break; 
	}
	return DefWindowProc(hwnd, msg, wParam, lParam);
}

int PopFileDlg(PTSTR pstrFileName, HWND hOwner)
{
	static OPENFILENAME ofn;
	static char *szFilter, *ext;

	ofn.nFilterIndex = 1;
	if (decode_mode == VC1)
		szFilter = \
			TEXT ("vc1\0*.vc1\0") \
			TEXT ("All Files (*.*)\0*.*\0");
	else if (decode_mode == AVC)
		szFilter = \
			TEXT ("264, h264, avc\0*.264;*.h264;*.avc\0") \
			TEXT ("All Files (*.*)\0*.*\0");
	else if (decode_mode == MP1)
		szFilter = \
			TEXT ("mp1, m1v\0*.mp1;*.m1v\0") \
			TEXT ("All Files (*.*)\0*.*\0");
	else if (decode_mode == MP2)
		szFilter = \
			TEXT ("mp2, m2v\0*.mp2;*.m2v\0") \
			TEXT ("All Files (*.*)\0*.*\0");

	ofn.lStructSize       = sizeof (OPENFILENAME);
	ofn.hwndOwner         = hOwner;
	ofn.hInstance         = NULL;
	ofn.lpstrFilter       = szFilter;
	ofn.nMaxFile          = 2048;
	ofn.nMaxFileTitle     = 2048;
	ofn.lpstrFile         = pstrFileName;
	ofn.lpstrInitialDir   = NULL;
	*ofn.lpstrFile = 0;
	ofn.Flags = OFN_HIDEREADONLY | OFN_FILEMUSTEXIST | OFN_ALLOWMULTISELECT | OFN_EXPLORER;
	return GetOpenFileName(&ofn);
}
