// (C) Copyright 2008 Donald A. Graft, All Rights Reserved

#include <windows.h>
#include <mmintrin.h>
#include "..\CUVIDServer\CUVIDServer.h"

unsigned char yuy2[2*MAX_WIDTH*MAX_HEIGHT];
unsigned char rgb24[3*MAX_WIDTH*MAX_HEIGHT];
static BITMAPINFOHEADER birgb;
unsigned char yuv2r[256][256][256];
unsigned char yuv2g[256][256][256];
unsigned char yuv2b[256][256][256];

void InitYUV2RGB(void)
{
	int y, u, v, r, g, b;

	for (y = 0; y < 256; y++)
	{
		for (u = 0; u < 256; u++)
		{
			for (v = 0; v < 256; v++)
			{
				r = (298 * (y - 16) + 409 * (v - 128) + 128) >> 8;
				if (r < 0)
					r = 0;
				else if (r > 255)
					r = 255;
				yuv2r[y][u][v] = r;

				g = (298 * (y - 16) - 100 * (u - 128) - 208 * (v - 128) + 128) >> 8;
				if (g < 0)
					g = 0;
				else if (g > 255)
					g = 255;
				yuv2g[y][u][v] = g;

				b = (298 * (y - 16) + 516 * (u - 128) + 128) >> 8;
				if (b < 0)
					b = 0;
				else if (b > 255)
					b = 255;
				yuv2b[y][u][v] = b;
			}
		}
	}
}

static void NV12toYUY2(const unsigned char *py8,
                       const unsigned char *puv8,
                       unsigned char *pyuy2_8,
                       int width, int height, int pitch, int interlaced)
{
    int w = (width+7) >> 3;
    const __m64 *py = (const __m64 *)py8;
    __m64 *pyuy2 = (__m64 *)pyuy2_8;
    
    for (int y=0; y<height; y++)
    {
        const __m64 *puv;
        
        if (interlaced)
            puv = (const __m64 *)(puv8+(((y/2)&~1)+(y&1))*pitch);
        else
            puv = (const __m64 *)(puv8+(y/2)*pitch);
        for (int x=0; x<w; x++)
        {
            __m64 r0 = py[x];       // yyyy
            __m64 r1 = puv[x];      // uvuv
            __m64 r2 = r0;          // yyyy
            r0 = _m_punpcklbw(r0, r1);  // yuyv
            r2 = _m_punpckhbw(r2, r1);  // yuyv
            pyuy2[x*2] = r0;
            pyuy2[x*2+1] = r2;
        }
        py += pitch/sizeof(__m64);
        pyuy2 += (width*2)/sizeof(__m64);
    }
    _m_empty();
}

void ShowFrame(int convert)
{
	unsigned char *p, *q, y, u, v;
	unsigned int h, w;
	extern HDC hDC;

	if (convert)
	{
		// Convert to RGB24 for display.
		NV12toYUY2(pMsg->nv12_frame,
					pMsg->nv12_frame + (pMsg->pitch * pMsg->height),
					yuy2,
					pMsg->width, pMsg->height,
					pMsg->pitch, pMsg->progressive);
		p = rgb24;
		q = yuy2 + 2 * pMsg->width * (pMsg->height - 1);
		for (h = 0; h < pMsg->height; h++)
		{
			for (w = 0; w < pMsg->width; w++)
			{
				y = q[2*w];
				u = q[2*(w&~1)+1];
				v = q[2*(w&~1)+3];
				p[3*w] = yuv2b[y][u][v];
				p[3*w+1] = yuv2g[y][u][v];
				p[3*w+2] = yuv2r[y][u][v];
			}
			p += 3 * pMsg->width;
			q -= 2 * pMsg->width;
		}
		birgb.biSize = sizeof(BITMAPINFOHEADER);
		birgb.biWidth = pMsg->width;
		birgb.biHeight = pMsg->height;
		birgb.biPlanes = 1;
		birgb.biBitCount = 24;
		birgb.biCompression = BI_RGB;
		birgb.biSizeImage = pMsg->width * pMsg->height * 3;
	}
	SetDIBitsToDevice(hDC, 0, 0, birgb.biWidth, birgb.biHeight, 0, 0, 0, birgb.biHeight,
					  rgb24, (LPBITMAPINFO)&birgb, DIB_RGB_COLORS);
}
